/*
 *  PTcommon.c
 *
 *  Many of the routines are based on the program PTStitcher by Helmut
 *  Dersch.
 * 
 *  Copyright Helmut Dersch and Daniel M. German
 *  
 *  Dec 2006
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  Author: Daniel M German dmgerman at uvic doooot ca
 * 
 */

#include "filter.h"
#include "PTcommon.h"
#include "ColourBrightness.h"

#include <assert.h>

//#include <stdio.h>
//#include <stdlib.h>
//#include <sys/types.h>
//#include <dirent.h>
//#include <unistd.h>
//#include <stdint.h>
//#include <math.h>



//declare functions
void getROI( TrformStr *TrPtr, aPrefs *aP, PTRect *ROIRect );

int ptQuietFlag = 0;

void InsertFileName( fullPath *fp, char *fname ){
#ifdef __Mac__
  strcpy( (char*)(fp->name), fname );
  c2pstr( (char*)(fp->name));
#else
  char *c = strrchr((char*)(fp->name), PATH_SEP);
  if(c != NULL) c++;
  else c = fp->name;
  strcpy( c, fname );
#endif
}       


int  CreatePSD(  fullPath *fullPathImages, int numberImages, fullPath* outputFileName)
{
  return 0;
}


static void ComputeStitchingMask8bits(Image *image)
{
  
  int column;
  int row;
  unsigned char *ptr;
  unsigned char *pixel;
  uint16_t  *ptrCounter;
  uint16_t count;
  
  // Use the GreenBlue pixel area is used to keep a counter of the
  // minimum distance (in pixels) away we are from the edges of the
  // mask (horizontal or vertical)
  
  // The algorithm is fairly simple:
  
  // For each column
  //   Process each row from top to down
  //     Set each pixel counter to the number of pixels from edge of the mask (from the left)
  //   Process each row from bottom to top
  // Set each pixel counter to the minimum between current counter and number of pixels
  // from edge (from the right)
  
  // for each row
  //   repeat the same algorithm (done per column)
  
  for (column = 0; column < image->width; column ++) {
    count = 0;
    // Point to the given column in row 0
    ptr = *image->data + column * 4;
    //    fprintf(stderr, "St1.1 Column[%d]\n", column);
    
    for (row = 0; row < image->height; row ++) {
      pixel = row * image->bytesPerLine + ptr;
      count = (*pixel == 0) ? 0 : count+1;
      
      ptrCounter = (uint16_t*)(pixel +2);
      *ptrCounter = count;
    }
    
    
    count = 0;
    row = image->height;
    
    while (--row  >= 0) {
      pixel = ptr + row + image->bytesPerLine;
      
      count = (*pixel == 0) ? 0 : count+1;
      
      ptrCounter = (uint16_t*)(pixel +2);
      if ( *ptrCounter < count)
        count = *ptrCounter; 
      else
        *ptrCounter  = count;
      
    } //while
    
    //    fprintf(stderr, "St1.5 Column[%d]\n", column);
    
    
  } //
  
  ///////////// row by row
  
  //  fprintf(stderr, "St2\n");
  
  for (row = 0; row < image->height; row ++) {
    count = image->width;
    ptr = row * image->bytesPerLine + *(image->data);
    
    // process from left to right
    for (column = 0; column < image->width; column ++ ) {
      pixel = ptr + 4 * column;
      count = (*pixel == 0) ? 0 : count+1;
      
      ptrCounter = (uint16_t*)(pixel +2);
      
      if (*ptrCounter > count)
        *ptrCounter = count;
    } // for column
    
    //-----------------------------;;
    
    
    //  fprintf(stderr, "St3\n");
    
    count = 0;
    column = image->width;
    
    while (--column >= 0) {
      pixel = ptr + column * 4;
      
      count = (*pixel == 0) ? 0 : count+1;
      
      ptrCounter =  (uint16_t*)(pixel +2);
      if (*ptrCounter < count)
        count = *ptrCounter;
      else
        *ptrCounter = count;
      
    } //    while (--column >= 0)
    
    
  } // end of for row
  
  
}

static void ComputeStitchingMask16bits(Image *image)
{
  
  int column;
  int row;
  unsigned char *ptr;
  uint16_t *pixel;
  uint16_t  *ptrCounter;
  uint16_t count;
  
  //  fprintf(stderr, "St1\n");
  
  for (column = 0; column < image->width; column ++) {
    
    count = 0;
    
    // Point to the given column in row 0
    
    ptr = *image->data + column * 8;
    
    //    fprintf(stderr, "St1.1 Column[%d]\n", column);
    
    for (row = 0; row < image->height; row ++) {
      
      //      fprintf(stderr, "St1.2 Column[%d] Row[%d]\n", column, row);
      
      pixel = (uint16_t*) (row * image->bytesPerLine + ptr);
      
      if (*pixel == 0) {
        
        count = 0;
        
      } else {
        
        count ++;
        
      }
      // Use the G pixel area to keep a count of how many pixels we have seen in 
      // the mask area.
      
      ptrCounter = pixel +2;
      *ptrCounter = count;
      
    } //     for (row = 0; row < image->heght; row ++) {
    
    
    //    fprintf(stderr, "St1.3 Column[%d]\n", column);
    
    count = 0;
    row = image->height;
    
    while (--row  >= 0) {
      
      //      fprintf(stderr, "St1.4 Column[%d] Row[%d]\n", column, row);
      
      pixel = (uint16_t*) (row * image->bytesPerLine + ptr);
      
      if ( *pixel == 0 ) {
        
        count = 0;
        
      } else {
        
        count ++;
        
      }
      
      ptrCounter = pixel +2;
      
      if ( *ptrCounter < count) {
        
        count = *ptrCounter; 
        
      } else {
        
        *ptrCounter  = count;
        
      }
      
      
    } //while
    
    //    fprintf(stderr, "St1.5 Column[%d]\n", column);
    
    
  } //
  
  ///////////// row by row
  
  //  fprintf(stderr, "St2\n");
  //  return; ///AAAAAAAAAAAAAAAA
  
  
  for (row = 0; row < image->height; row ++) {
    
    count = image->width;
    
    ptr = row * image->bytesPerLine + *(image->data);
    
    for (column = 0; column < image->width; column ++ ) {
      
      
      pixel = (uint16_t*)(ptr + 4 * 2 * column);
      
      if ( *pixel == 0 ) {
        
        count = 0;
        
      } else {
        
        count ++;
        
      }
      
      ptrCounter = pixel +2;
      
      if (*ptrCounter < count) {
        
        // count = *ptrCounter; AAAAAAAAAAAAA
        
      } else {
        
        *ptrCounter = count;
        
      } //
      
    } // for column
    
    //-----------------------------;;
    
    
    for (column = 0; column < image->width; column ++ ) {
      
      
      pixel = (uint16_t*)(ptr + 4 * 2 * column);
      
      if ( *pixel == 0 ) {
        count = 0;
      } else {
        count ++;
      }
      
      ptrCounter = pixel +2;
      
      if (*ptrCounter < count) {
        // count = *ptrCounter; AAAAAAAAAAAAAAA
      } else {
        *ptrCounter = count;
      }
    } // for
    
    //---------------------------------------;
    
    //  fprintf(stderr, "St3\n");
    
    count = image->width;
    column = image->width;
    
    while (--column >= 0) {
      
      pixel = (uint16_t*)(ptr + 4 * 2 * column);
      
      if (0 == *pixel ) {
        count = 0;
      } else {
        count ++;
      }
      
      ptrCounter =  pixel +2;
      
      if (*ptrCounter < count) {
        count = *ptrCounter;
      } else {
        *ptrCounter = count;
      }
    } //    while (--column >= 0) {
    
    //--------------------------------;
    column = image->width;
    
    //  fprintf(stderr, "St4\n");
    
    while (--column >= 0) {
      
      pixel = (uint16_t*)(ptr + 4 * 2 * column);
      
      
      if ( *pixel == 0 ) {
        count = 0;
      } else {
        count ++;
      }
      
      ptrCounter =  pixel +2;
      
      if (*ptrCounter < count) {
        
        count = *ptrCounter ;
        
      } else {
        *ptrCounter = count;
      }
    } // end of while
    
  } // end of for row
}


static void ComputeStitchingMaskMap(Image *image)
{
  if ( image->bitsPerPixel == 32 ) {
    ComputeStitchingMask8bits(image);
    return;
  } else  if ( image->bitsPerPixel == 64 ) {
    ComputeStitchingMask16bits(image);
    return;
  }
  fprintf(stderr, "Masking not supported for this image type (%d bitsPerPixel)\n", (int)image->bitsPerPixel);
  exit(1);
}




//
// Compute the map of the stitching mask and create a file with it.
// The stitching mask will be contained in the GB channels (this is,
// the 16 bits corresponding to the G and B channel will contain a uint16_t that
// contains, for that particular point, the stitching mask.
//
int CreateMaskMapFiles(fullPath *inputFiles, fullPath *maskFiles, int numberImages)
{
}





/*
  MRDL: ReplaceAlphaChannels doesn't seem to be used?
  static int  ReplaceAlphaChannels(fullPath *inputImagesNames, fullPath *masksNames, fullPath *outputNames, int numberImages,
  pt_tiff_parms *imageDefaultParms)
  {
  int index;
  unsigned char *imageRowBuffer = NULL;
  unsigned char *maskRowBuffer = NULL;
  char tempString[MAX_PATH_LENGTH];
  int row;
  int j;
  
  TIFF *imageFile;
  TIFF *outputFile;
  TIFF *maskFile;
  
  if ((imageRowBuffer = calloc(imageDefaultParms->bytesPerLine,1 )) == NULL ||
  (maskRowBuffer = calloc(imageDefaultParms->bytesPerLine,1 )) == NULL ) {
  PrintError("Not enough memory");
  return -1;
  }
  
  // The algorith is fairly simple. 
  
  // For each image
  //   For each row in that image
  //     Read row of image
  //     Read row of mask
  //     Replace alpha channel in image with masks alpha channel
  //     Write row
  
  if (!ptQuietFlag) {
  Progress(_initProgress, "Inserting Alpha Channel");
  }
  
  for (index = 0; index < numberImages; index ++) { //
  
  // Do one file at a time
  
  if ( ptQuietFlag == 0 ) {
  sprintf(tempString, "%d", index * 100/ numberImages);
  if (Progress(_setProgress, tempString) == 0) {
  return 0;
  }
  }
  
  // Open input image 
  if ( (imageFile = OpenTiffFromFullPath(&inputImagesNames[index], "r")) == NULL) {
  PrintError("Could not open TIFF-file");
  return 0;
  }
  
  // Open mask file
  if ( (maskFile = OpenTiffFromFullPath(&masksNames[index], "r")) == NULL) {
  PrintError("Could not open mask file");
  return 0;
  }
  
  // Create output file
  if ((outputFile = OpenTiffFromFullPath(&outputNames[index], "w")) == NULL) {
  PrintError("Could not create TIFF-file");
  return 0;
  }
  if (!TiffSetImageParameters(outputFile,imageDefaultParms)) {
  PrintError("Unable to initialize TIFF file\n");
  return 0;
  }
  
  // start processing each row
  
  for (row = 0; row < imageDefaultParms->imageLength; row ++) {
  
  unsigned char *source;
  unsigned char *destination;
  
  TIFFReadScanline(imageFile, imageRowBuffer, row, 0);
  
  TIFFReadScanline(maskFile, maskRowBuffer, row, 0);
  
  // Depending on the type of image, process...
  
  if ( imageDefaultParms->bitsPerPixel == 32 ) {
  
  destination = imageRowBuffer + 3;
  source = maskRowBuffer + 3;
  
  for (j = 0; j < imageDefaultParms->imageWidth; j ++ ) {
  *destination = *source;
  destination +=4;
  source +=4;
  } 
  
  } else {
  
  destination = imageRowBuffer + 6;
  source = maskRowBuffer + 6;
  
  for (j = 0; j < imageDefaultParms->imageWidth ; j ++ ) {
  *destination = *source;
  
  source += 8;
  destination += 8;
  
  } // for j
  
  } // end of if
  
  if (TIFFWriteScanline(outputFile, imageRowBuffer, row, 0) != 1) {
  PrintError("Unable to write scan line\n");
  TIFFClose(imageFile);
  TIFFClose(maskFile);
  TIFFClose(outputFile);
  return 0;
  }
  } // for 
  
  TIFFClose(imageFile);
  TIFFClose(maskFile);
  TIFFClose(outputFile);
  
  } // for index
  
  if (!ptQuietFlag) 
  Progress(_disposeProgress, tempString);
  
  
  free(imageRowBuffer);
  free(maskRowBuffer);
  
  return 1;
  
  }
*/

static int  ReplaceAlphaChannel(fullPath *inputImage, fullPath *mask, fullPath *output)
{
}


#ifdef __Win__
//void InsertFileName( fullPath *fp, char *fname ){
// char *c = strrchr((char*)(fp->name), PATH_SEP);
// if(c != NULL) c++;
// else c = fp->name;
// strcpy( c, fname );
//}   
#endif


static int Create_LP_ivr(Image *image, fullPath* fullPathImage)
{
  fprintf(stderr,"Create_LP_ivr this function is not implemented yet\n");
  exit(1);
}

static int Unknown01(Image *image, fullPath* fullPathImage)
{
  fprintf(stderr,"Unknown01 this function is not implemented yet\n");
  exit(1);
}

static int Unknown02(Image *image, fullPath* fullPathImage)
{
  fprintf(stderr,"Unknown02 this function is not implemented yet\n");
  exit(1);
}

static int Unknown03(Image *image, fullPath* fullPathImage)
{
  fprintf(stderr,"Unknown03 this function is not implemented yet\n");
  exit(1);
}

static int Unknown04(Image *image, fullPath* fullPathImage)
{
  fprintf(stderr,"Unknown04 this function is not implemented yet\n");
  exit(1);
}

static int Unknown05(Image *image, fullPath* fullPathImage)
{
  fprintf(stderr,"Unknown05 this function is not implemented yet\n");
  exit(1);
}

static int Create_QTVR(Image *image, fullPath* fullPathImage)
{
  fprintf(stderr,"Create QTVR is not implemented yet\n");
  exit(1);
}

static void ARGtoRGBAImage(Image *im)
{

}



void Clear_Area_Outside_Selected_Region(Image *image)
{
  // This function clears (i.e. sets to zero) the area outside the 
  // selection region 
  
  int right;    
  int left;
  int bottom;
  int top;
  //  int width;
  //  int var24;
  int bytesPerPixel;  // 32
  unsigned char *dataPtr;
  unsigned char *pixelPtr;
  
  int currentRow;
  int currentColumn;
  
  // Only works for 8/16 bit per channel (32/64 bits per pixel) images
  assert(image->bitsPerPixel == 32 || image->bitsPerPixel == 64);
  
  top = image->selection.top;
  bottom = image->selection.bottom;
  left = image->selection.left;
  right = image->selection.right;
  
  if ( bottom == 0 )
    bottom = image->height;
  
  if ( right == 0 )
    right = image->width;
  
  if ( image -> format == _fisheye_circ) {
    PrintError("Not implemented yet");
    exit(1);
  }
  
  if ( image->bitsPerPixel == 32 ) {
    bytesPerPixel  = 4;
  } else if (image->bitsPerPixel == 64) {
    bytesPerPixel  = 8;    
  } else {
    assert(0); // it should not reach here 
    exit(0);
  }
  
  // Clear the area at above the image
  dataPtr = *(image->data);
  
  for (currentRow =0; currentRow < top; currentRow++ ) {
    pixelPtr = dataPtr;
    
    for (currentColumn = 0;  currentColumn < image->width ;   currentColumn++) {
      assert(sizeof(int) == bytesPerPixel);
      memset(pixelPtr, 0, bytesPerPixel);
      pixelPtr+=bytesPerPixel;
    }
    
    dataPtr += image->bytesPerLine;
  }
  
  // Clear area below the picture
  dataPtr = bottom * image->bytesPerLine + *(image->data);
  
  for (currentRow=bottom    ;  currentRow < image->height ; currentRow++) {
    pixelPtr = dataPtr;
    for (currentColumn = 0; currentColumn < image->width ; currentColumn++) {
      memset(pixelPtr, 0, bytesPerPixel);
      pixelPtr += bytesPerPixel;
    }
    
    dataPtr += image->bytesPerLine;
    
  } //  for (    ;  %currentColumn < image->width ; currentColumn++,pixelPtr += bytesPerPixel) {
  
  
  /* Clear the area to the left of the picture */
  
  dataPtr = *(image->data);
  for (   currentRow = 0 ; currentRow < image->height;   currentRow++) {
    
    pixelPtr = dataPtr;      
    for (currentColumn = 0 ; currentColumn < left ; currentColumn++) {
      memset(pixelPtr, 0, bytesPerPixel);
      pixelPtr += bytesPerPixel;
    }
    
    dataPtr += image->bytesPerLine;
  }
  
  /* Clear the area to the right of the picture */
  
  dataPtr = *(image->data);
  
  for (currentRow = 0; currentRow < image->height; currentRow++ ) {
    
    pixelPtr = dataPtr + bytesPerPixel * right;
    
    for (currentColumn=right ;currentColumn < image->width; currentColumn++) {
      
      memset(pixelPtr, 0, bytesPerPixel);
      
      pixelPtr += bytesPerPixel;
      
    }
    
    dataPtr += image->bytesPerLine;
    
  }
  
  return;
  
  
#ifdef not_implemented_yet
  
  
  //  THIS IS the code for fisheye_circular 24 bits, but I don't understand it yet.
  
  pixelPtr = right;
  currentColumn = left;
  
  eax = left + right;
  
  
  var20 = (left + right)/2;
  var24 = (top + bottom) /2;
  assert(left >= right);
  temp = (left - right ) /2;
  tmpStr28 = temp * temp;
  
  
  dataPtr =  *(image->data);
  
  for (currentRow = 0 ; currentRow < image->height ; currentRow++) {
    
    currentColumn = 0;
    pixelPtr = dataPtr;
    
    if ( currentColumn < image->width ) {
      
      
      temp = currentRow - var24;
      
      var36 = temp * temp;
      
      do { 
        
        temp = currentColumn - var20;
        temp = temp * temp + var36;
        
        if ( %eax > tmpStr28 ) {
          
          *pixelPtr = 0; //moves only 1 byte
        }
        
        currentColumn++;
        pixelPtr += bytesPerPixel;
        
      } while ( currentColumn < image->width );
      
    } //    if ( currentColumn < image->width ) {
    
    
    dataPtr += image-> bytesPerLine;
    
    
  } //for ( ; currentRow < image->height ; ) {
  
  return;
} //if ( image->bitsPerPixel == $0x20 ) {
#endif



#ifdef not_implemented_yet

THIS IS the code for fisheye_circular 64 bits


var20 = (left + right) /2;

var24 = (top + bottom)/2;

currentColumn = (left - right )/2;

tmpStr28 = currentColumn * currentColumn;

dataPtr = *(image->data);

for (  currentRow = 0; ; currentRow < image->height ; ++currentRow) {
  //if ( currentRow >= image->height )
  //              return;
  
  currentColumn = 0;
  pixelPtr = dataPtr;
  
  if ( currentColumn < image->width ) {
    
    var40 = (currentRow * var24) * (currentRow * var24);
    
    do {
      
      eax = currentColumn * currentColumn + var40;
      
      if ( %eax > tmpStr28 ) {
        *pixelPtr = 0; // Again CAREFUL, moves 8 bytes
      }
      
      currentColumn ++;
      pixelPtr = bytesPerPixel;
      
    } while (currentColumn < image->width );
    //if ( currentColumn < image->width )
    //       
    
  } //if ( currentColumn < image->width ) {
  
  dataPtr +=image->bytesPerLine;
  
 } //for ( ; %currentRow < image->height ; ) 
#endif

return;


}


void Unknown09(Image *currentImagePtr)
{
  // NEEDED
  fprintf(stderr,"Unknown09 this function is not implemented yet\n");
  exit(1);
}


/**
 * This function computes the minimal rectangle needed to encompass
 * the region of the output image (TrPtr->dest) that will be populated with 
 * data from the input image (TrPtr->src) using the options specified
 * in aP.  The ROIRect is populated with the left/right/top/bottom values
 * that define this ROI within the output image
 */
void getROI( TrformStr *TrPtr, aPrefs *aP, PTRect *ROIRect )
{
  struct   MakeParams  mpinv;
  fDesc    invstack[15], finvD;
  int   color               = 0;
  
  int             x, y, x_jump;
  double             x_d, y_d;  // Cartesian Coordinates of point in source (i.e. input) image
  double           Dx, Dy;      // Coordinates of corresponding point in destination (i.e. output) image
  
  double             w2   = (double) TrPtr->dest->width  / 2.0 - 0.5;   //half destination image width
  double             h2   = (double) TrPtr->dest->height / 2.0 - 0.5;   //half destination image height
  double             sw2 = (double) TrPtr->src->width   / 2.0 - 0.5;   //half source image width
  double             sh2 = (double) TrPtr->src->height  / 2.0 - 0.5;   //half source image height
  
  //Set initial values for ROI to be adjusted during this function
  ROIRect->left         = TrPtr->dest->width-1;
  ROIRect->right        = 0;
  ROIRect->top          = TrPtr->dest->height-1; 
  ROIRect->bottom       = 0;
  
  //The "forward" transform (although not used here) allows us to map pixel
  //coordinates in the output image to their location in the source image.
  //SetMakeParams( stack, &mp, &(aP->im) , &(aP->pano), color );
  //fD.func = execute_stack; fD.param = stack;
  
  //The "inverse" transform allows us to map pixel coordinates in each source image
  //to their location in the output image.
  SetInvMakeParams( invstack, &mpinv, &(aP->im) , &(aP->pano), color );   
  finvD.func = execute_stack_new; 
  finvD.param = invstack;
  
  //iterate over edges of input image and compute left/right/top/bottom-most coordinate
  //in output image
  //For equirectangular output projection covering 360/180, iterating over the 
  //edges of each input image isn't sufficient to determine ROI because an 
  //an interior point in an input image can be at the edge of ROI.  More research 
  //needed here, but for now include some representative interior points as well.
  for (y = 0; y <= TrPtr->src->height; y += 1) {
    
    x_jump = (y==0 || y==TrPtr->src->height) ? 1 : TrPtr->src->width/2;
    
    for (x = 0; x <= TrPtr->src->width; x += x_jump) {
      //convert source coordinates to cartesian coordinates (i.e. origin at center of image)
      x_d = (double) x - sw2 ;
      y_d = (double) y - sh2 ;
      
      //Map the source image cartesian coordinate to the destination image cartesian coordinate
      finvD.func( x_d, y_d, &Dx, &Dy, finvD.param);
      
      //Convert destination cartesian coordinate back to destination "screen" coordinates (i.e. origin at top left of image)
      Dx += w2;
      Dy += h2;
      
      //printf("  IN: %d,%d -> OUT: %d, %d\n", x, y, (int)Dx, (int)Dy);
      
      //Expand ROI if necessary
      if ((int)Dx < ROIRect->left) ROIRect->left = (int)Dx;
      if ((int)Dx > ROIRect->right) ROIRect->right = (int)Dx;
      if ((int)Dy < ROIRect->top) ROIRect->top = (int)Dy;
      if ((int)Dy > ROIRect->bottom) ROIRect->bottom = (int)Dy;
    }
  }
  
  //Reduce ROI if it extends beyond boundaries of final panorama region
  if (ROIRect->left    < 0) ROIRect->left =0;
  if (ROIRect->top     < 0) ROIRect->top  =0;
  
  if (ROIRect->right   > (TrPtr->dest->width-1))  ROIRect->right    = TrPtr->dest->width-1;  
  if (ROIRect->bottom  > (TrPtr->dest->height-1)) ROIRect->bottom   = TrPtr->dest->height-1;
  
  //printf("ROI: %d,%d - %d, %d\n", ROIRect->left, ROIRect->top, ROIRect->right, ROIRect->bottom);
}




int CreatePanorama(fullPath ptrImageFileNames[], int counterImageFiles, fullPath *panoFileName, fullPath *scriptFileName)
{
 
}

/*
 * replace the extension in filename with
 extension
*/
void ReplaceExt(char* filename, char *extension)
{
  char *temp;
  temp = strrchr(filename, '.');
  if (temp != NULL) {
    strcpy(temp, extension);
  } else {
    strcat(filename, extension);
  }
  return;
}
