#include <stdio.h>
#include <locale.h>

#include "filter.h"

extern AlignInfo	*g;	// This struct holds all informations for the optimization

#ifdef _WINDOWS_
#include <Windows.h>
#define filenamechartype wchar_t
#else
#define filenamechartype char
#endif

int 			CheckParams( AlignInfo *g );
int main (int argc, const char * argv[]);
void RunOptimizer(const char *scriptfilename);
void fileToString(const filenamechartype *filename, char **string);
void WriteResults_Modified( char* script, const filenamechartype *sfile,  AlignInfo *g, double ds( int i), int launch);
void stringToFile(const filenamechartype *filename, char *string);

#define LINE_LENGTH         65536

int main (int argc, const char * argv[]) {
  if(argc == 2)
  {
    RunOptimizer(argv[1]);
  }
  else
  {
    fprintf(stderr,"Usage: ptoptimizer <name of script file>\n");
    return 1;
  }
}

#ifdef _WINDOWS_
int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PWSTR pCmdLine, int nCmdShow)
{
  LPWSTR *argv;
  int argc;
  argv = CommandLineToArgvW(GetCommandLineW(), &argc);
  if(argc == 2)
  {
    RunOptimizer(argv[1]);
  }
  else
  {
    fprintf(stderr,"Usage: ptoptimizer <name of script file>\n");
    return 1;
  }
}
#endif

void RunOptimizer(const filenamechartype *scriptfilename)
{
  char*				script;
  OptInfo				opt;
  AlignInfo			ainf;

  fileToString(scriptfilename, &script);
  if (ParseScript( script, &ainf ) == 0)
  {
    if( CheckParams( &ainf ) == 0 ) 				// and it seems to make sense
    {
      ainf.fcn	= fcnPano;
      
      SetGlobalPtr( &ainf ); 
      
      opt.numVars 		= g->numParam;
      opt.numData 		= g->numPts;
      opt.SetVarsToX		= SetLMParams;
      opt.SetXToVars		= SetAlignParams;
      opt.fcn				= g->fcn;
      *opt.message		= 0;

      RunLMOptimizer( &opt );
      g->data				= opt.message;
      WriteResults_Modified( script, scriptfilename, g, distSquared ,
                  0 );
    }
    DisposeAlignInfo( &ainf );					// These were allocated by 'ParseScript()'
  }
  free( script );
}

void fileToString(const filenamechartype *filename, char **string)
{
#ifdef _WINDOWS_
  FILE *fhandle=_wfopen(filename,L"rb");
#else
  FILE *fhandle=fopen(filename,"rb");
#endif
  long filesize;
  if(fhandle)
  {
    fseek(fhandle,0,SEEK_END);
    filesize=ftell(fhandle);
    rewind(fhandle);
    
    *string=malloc(filesize+1);
    if(*string)
    {
      int bytesread=fread(*string,1,filesize,fhandle);
      if(bytesread != filesize)
      {
        fprintf(stderr,"File read error");
        exit(1);
      }
      (*string)[filesize]=0;
    }
    else
    {
      fprintf(stderr,"Out of memory");
      exit(1);
    }
    fclose(fhandle);
  }
  else {
    fprintf(stderr,"Could not read file: %s\n",filename);
    exit(1);
  }
}

void stringToFile(const filenamechartype *filename, char *string)
{
#ifdef _WINDOWS_
  FILE *fhandle=_wfopen(filename,L"wb");
#else
  FILE *fhandle=fopen(filename,"wb");
#endif
  if(fhandle)
  {
    size_t numbytes=strlen(string);
    int byteswritten=fwrite(string,1,numbytes,fhandle);
    if(byteswritten != numbytes)
    {
      fprintf(stderr,"File read error");
      exit(1);
    }
    fclose(fhandle);    
  }
}

void WriteResults_Modified( char* script, const filenamechartype *sfile,  AlignInfo *g, double ds( int i), int launch)
{
	char 		*res, line[LINE_LENGTH], cmd[LINE_LENGTH];
	int 		optHfov;		 // Has hfov being optimized?
	int			opta,optb,optc;  // same for a,b,c - the polynomial corrections
	int         format;
	int i;

	setlocale(LC_ALL, "C");
		
	res = malloc( strlen(script) + g->numIm * 600 + g->numPts * 200 + 10000 ); // Do we ever need more?
  res[0]=0;
	if( res == NULL )
	{
    fprintf(stderr,"Out of memory");
    exit(1);
	}
	strcat( res, script );
	sprintf( line, "\n*\n\n" ); strcat( res, line );
	sprintf( line, "# ====================================================================\n");strcat( res, line );
	sprintf( line, "# Output  generated by Panorama Tools\n\n" );strcat( res, line );
	if( g->data != NULL )
		strcat( res, (char*) g->data );
	
		
	sprintf( line, "\n# Panorama description\n" );strcat( res, line );
	switch( g->pano.format )
	{
                case _mercator:		        format = 4; break;
                case _equirectangular:		format = 2; break;
		case _rectilinear:			format = 0; break;
		case _panorama:				format = 1; break;
		default:			format = -1; break;
	}
		
	sprintf( line, "# p f%d w"FMT_INT32" h"FMT_INT32" v%g n\"%s\"\n\n", format, g->pano.width, g->pano.height, g->pano.hfov, g->pano.name );strcat( res, line );
	
	
	
	sprintf( line, "# Parameters for Each Input Image:\n" );strcat( res, line );
	sprintf( line, "# (*) - optimized         (p) - preset \n\n");strcat( res, line );
			
	for( i=0; i<g->numIm; i++ )
	{
		switch( g->im[i].format )
		{
			case _rectilinear 		: format = 0; break;
			case _panorama    		: format = 1; break;
			case _fisheye_circ		: format = 2; break;
			case _fisheye_ff  		: format = 3; break;
			case _equirectangular 	: format = 4; break;
			default: break;
		}
		sprintf( cmd, "o f%d ", format );

		if( g->opt[i].hfov == 1 ||
			( g->opt[i].hfov > 1 &&  g->opt[g->opt[i].hfov-2].hfov == 1 ))
			optHfov = 1;
		else
			optHfov = 0;

		if( g->opt[i].a == 1 ||
			( g->opt[i].a > 1 &&  g->opt[g->opt[i].a-2].a == 1 ))
			opta = 1;
		else
			opta = 0;
					
		if( g->opt[i].b == 1 ||
			( g->opt[i].b > 1 &&  g->opt[g->opt[i].b-2].b == 1 ))
			optb = 1;
		else
			optb = 0;
					
		if( g->opt[i].c == 1 ||
			( g->opt[i].c > 1 &&  g->opt[g->opt[i].c-2].c == 1 ))
			optc = 1;
		else
			optc = 0;

		sprintf( line, "# Image No %d:\n", i );strcat( res, line );
		sprintf( line, "# Yaw:	%g deg (%c)	Pitch:	%g deg (%c) \n# Roll:	%g deg (%c)	HFov:	%g deg (%c)\n# Polynomial Coefficients: a   %f (%c); b   %f (%c); c   %f (%c)\n# Horizontal Shift: %f (%c)   Vertical Shift:  %f (%c)\n",
					g->im[i].yaw, ( g->opt[i].yaw ? '*' : 'p' ),
					g->im[i].pitch,  ( g->opt[i].pitch  ? '*' : 'p' ),
					g->im[i].roll, ( g->opt[i].roll ? '*' : 'p' ),
					g->im[i].hfov,  (              optHfov  ? '*' : 'p' ),
					g->im[i].cP.radial_params[0][3], (opta ? '*':'p'),
					g->im[i].cP.radial_params[0][2], (optb ? '*':'p'),
					g->im[i].cP.radial_params[0][1], (optc ? '*':'p'),
					g->im[i].cP.horizontal_params[0], (g->opt[i].d ? '*':'p'),
					g->im[i].cP.vertical_params[0], (g->opt[i].e ? '*':'p')
					);
		strcat( res, line );
		sprintf( line, "r%g p%g y%g v%g ", g->im[i].roll,g->im[i].pitch,g->im[i].yaw,g->im[i].hfov);
		strcat( cmd, line );
		if( g->im[i].cP.radial )
		{
			sprintf( line, "a%f b%f c%f ", g->im[i].cP.radial_params[0][3], g->im[i].cP.radial_params[0][2], g->im[i].cP.radial_params[0][1]);
			strcat( cmd, line );
		}
		if( g->im[i].cP.shear )
		{
			sprintf( line, "g%f t%f ", g->im[i].cP.shear_x, g->im[i].cP.shear_y);
			strcat( cmd, line );
		}
		if( g->im[i].cP.cutFrame &&
		 !( g->im[i].selection.bottom != 0 || g->im[i].selection.right != 0 )) // g->im[i].format != _fisheye_circ && g->im[i].cP.cutFrame )
		{
			if( g->im[i].cP.frame != 0 )
			{
				sprintf( line, "m%d ",g->im[i].cP.frame );
			}
			else
			{
				sprintf( line, "mx%d my%d ",g->im[i].cP.fwidth, g->im[i].cP.fheight );
			}
			strcat( cmd, line );
		}

//		sprintf( line, "d%f e%f ", g->prefs[i]->c_prefs.horizontal_params[0], g->prefs[i]->c_prefs.vertical_params[0]);
//		strcat( cmd, line );
#if 0
		if( g->prefs[i]->colCorrect )
		{
			sprintf( line, "t%f,%f,%f,%f,%f,%f ", g->prefs[i]->ColCoeff[0][0],
														g->prefs[i]->ColCoeff[0][1],
														g->prefs[i]->ColCoeff[1][0],
														g->prefs[i]->ColCoeff[1][1],
														g->prefs[i]->ColCoeff[2][0],
														g->prefs[i]->ColCoeff[2][1]);
			strcat( cmd, line );
		}
#endif
		if( g->im[i].cP.horizontal )
		{
			sprintf( line, "d%f ",g->im[i].cP.horizontal_params[0]); 
			strcat( cmd, line );
		}
		if( g->im[i].cP.vertical )
		{
			sprintf( line, "e%f ",g->im[i].cP.vertical_params[0]);
			strcat( cmd, line );
		}
		if( g->im[i].cP.correction_mode & correction_mode_morph )
		{
			strcat( cmd, "o " );
		}
		if( g->im[i].selection.bottom != 0 || g->im[i].selection.right != 0 ){
			if( g->im[i].cP.cutFrame ){
				sprintf( line, " C"FMT_INT32","FMT_INT32","FMT_INT32","FMT_INT32" ",g->im[i].selection.left, g->im[i].selection.right,
						       g->im[i].selection.top, g->im[i].selection.bottom );
			}else{
				sprintf( line, " S"FMT_INT32","FMT_INT32","FMT_INT32","FMT_INT32" ",g->im[i].selection.left, g->im[i].selection.right,
						       g->im[i].selection.top, g->im[i].selection.bottom );
			}
			strcat( cmd, line );
		}


		sprintf(line, "u%d ", g->st.feather);
		strcat( cmd, line );
#if 0
		// Print filename
		if( g->im[i].name != NULL && strlen( g->im[i].name != 0 ) ){
			sprintf(line, " n\"%s\" ", g->im[i].name);
			strcat( cmd, line );
		}
	
#endif
		// Add command for stitcher, depending on g->st
		// If '-' option has been set (destName != 0), do nothing
		// else add stitch commands
		if( !*(g->st.destName) )
		{
			
			if( i== 0 )
			{
				sprintf(line, "-%s ", g->st.srcName);
				strcat( cmd, line );
			}
			else if( i == g->numIm-1 )
			{
				sprintf(line, "+%s ", g->st.srcName);
				strcat( cmd, line );
			}
			else
			{
				sprintf(line, "+%s -%s ", g->st.srcName, g->st.srcName);
				strcat( cmd, line );
			}
			// Print Feather commands
		}
		
				
		
		if( opta || optb || optc )
		{
			sprintf(line,"# 4th polynomial coefficient: %g\n", 1.0 - ( g->im[i].cP.radial_params[0][3]
																 + g->im[i].cP.radial_params[0][2]
																 + g->im[i].cP.radial_params[0][1] ) );
			strcat( res, line );
		}
		sprintf(line, "# Command for Panorama Creation: \n%s\n", cmd );
		strcat( res, line );
		sprintf(line, "\n");strcat( res, line );
	} // numIm

	if( g->numPts > 0 ) // Display Controlpoint distances
	{
		sprintf(line, "\n");strcat( res, line );
		sprintf( line, "# ==========================================================================\n");strcat( res, line );
		sprintf( line, "# Control Points: Distance between desired and fitted Position (in \"Pixels\")\n\n");strcat( res, line );
			
		for( i=0; i<g->numPts; i++ )
		{
			sprintf( line, "# Control Point No %d:  %g\n", i , sqrt ( ds(i)) );strcat( res, line );
		}
		// Print optimum positions for points in panorama
		for( i=0; i<g->numPts; i++ )
		{
			double x[2],y[2],xd,yd;
			
			// write only normal control points, not horizontals, verticals, or lines
			if (g->cpt[i].type == 0)
			{
				GetControlPointCoordinates(i, x, y, g );
				// Write only points inside image
				if( x[0] >= 0.0 && x[0] < g->pano.width && x[1] >= 0.0 && x[1] < g->pano.width &&
					y[0] >= 0.0 && y[0] < g->pano.height && y[1] >= 0.0 && y[1] < g->pano.height)
				{
					xd = (x[0]+x[1]) / 2.0;
					yd = (y[0]+y[1]) / 2.0;
					sprintf( line, "C i%d  x%g y%g X%g Y%g\n", g->cpt[i].num[0] , x[0], y[0], xd, yd );strcat( res, line );
					sprintf( line, "C i%d  x%g y%g X%g Y%g\n", g->cpt[i].num[1] , x[1], y[1], xd, yd );strcat( res, line );
				}
			}
		}
	}
  stringToFile(sfile,res);
  free(res);
	return;
}

